#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-07-16
# File  : bgp_topology/lib/bgp_topology.py

# 2024-07-20: moved to lib -> is now an active check

__AUTHOR__ = 'thl-cmk[at]outlook[dot]com'
__URL__ = 'https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/bgp_topology'
__USAGE__ = '/local/lib/python3/cmk_addons/plugins/bgp_topology/libexec/check_bgp_topology --make-default'
__VERSION__ = '0.0.1-20240721'

from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from collections.abc import MutableMapping, MutableSequence, Sequence
from dataclasses import dataclass
from time import time_ns

from cmk.agent_based.v2 import render
from cmk_addons.plugins.bgp_topology.constants import (
    ARG_PARSER_ANCHOR,
    ARG_PARSER_ANCHOR_AS,
    ARG_PARSER_ANCHOR_BOTH,
    ARG_PARSER_ANCHOR_ID,
    ARG_PARSER_EMBLEM_AS,
    ARG_PARSER_EMBLEM_ID,
    ARG_PARSER_HOST,
    ARG_PARSER_MAKE_DEFAULT,
    ARG_PARSER_NONE,
    ARG_PARSER_SITES_EXCLUDE,
    ARG_PARSER_SITES_INCLUDE,
    BGP_PEER_LOCAL_ADDR,
    BGP_PEER_LOCAL_AS,
    BGP_PEER_LOCAL_ID,
    BGP_PEER_REMOTE_ADDR,
    BGP_PEER_REMOTE_AS,
    BGP_PEER_REMOTE_ID,
    BGP_PEER_STATE,
    BGP_PEER_UPTIME,
    EMBLEM_BGP_AS,
    EMBLEM_BGP_ID,
    METRIC_TIME_TAKEN,
    PARAM_BGP_EXT_ANCHOR_AS,
    PARAM_BGP_EXT_ANCHOR_BOTH,
    PARAM_BGP_EXT_ANCHOR_ID,
    TOPOLOGY_NAME,
)
from cmk_addons.plugins.bgp_topology.lib.utils import (
    BASE_TOPO_PATH,
    LiveStatusConnection,
    Metric,
    OMD_ROOT,
    TopoConnections,
    TopoObjects,
    add_dummy_topologies,
    get_anchor,
    get_bgp_peer_clean_key,
    get_bgp_peer_clean_value,
    get_emblem,
    get_service,
    make_topo_default,
    save_topology,
)


class Params(Namespace):
    bgp_anchor: str
    make_default: bool = False
    bgp_emblem_as: str = EMBLEM_BGP_AS
    bgp_emblem_id: str = EMBLEM_BGP_ID
    include_sites: Sequence[str] | None = None
    exclude_sites: Sequence[str] | None = None
    host: str = ''


@dataclass(frozen=True)
class BgpPeer:
    host: str
    service: str
    state: int
    state_str: str
    remote_address: str
    uptime: str | None = None
    local_as: str | None = None
    local_address: str | None = None
    local_id: str | None = None
    remote_as: str | None = None
    remote_id: str | None = None

    @classmethod
    def parse(cls, host: str, state: int, servie: str, raw_peer_data: str):
        raw_peer_attributes: Sequence[str] = raw_peer_data.split('\\n')

        bgp_peer: MutableMapping[str, str] = {}
        for entry in raw_peer_attributes:
            try:
                key, value = entry.split(':', 1)
            except ValueError:
                key = None
                value = None
            key: str | None = get_bgp_peer_clean_key(key)
            value: str | None = get_bgp_peer_clean_value(key, value)
            if key is not None and value is not None:
                bgp_peer[key] = value

        return cls(
            host=host,
            local_address=bgp_peer.get(BGP_PEER_LOCAL_ADDR),
            local_as=bgp_peer.get(BGP_PEER_LOCAL_AS),
            local_id=bgp_peer.get(BGP_PEER_LOCAL_ID),
            remote_address=bgp_peer.get(BGP_PEER_REMOTE_ADDR),
            remote_as=bgp_peer.get(BGP_PEER_REMOTE_AS),
            remote_id=bgp_peer.get(BGP_PEER_REMOTE_ID),
            service=servie,
            state=state,
            state_str=bgp_peer.get(BGP_PEER_STATE),
            uptime=bgp_peer.get(BGP_PEER_UPTIME),
        )


def create_bgp_topology(params: Params | None) -> int:
    start_time = time_ns()

    summary: list[str] = []
    details: list[str] = []

    objects = TopoObjects()
    connections = TopoConnections()
    sub_directory = TOPOLOGY_NAME

    ls_connection = LiveStatusConnection()
    if params.include_sites is not None:
        ls_connection.filter_sites(include=True, sites=params.include_sites)
        sites_str: str = ', '.join(params.include_sites)
        details.append(f'Site(s) included: {sites_str}')
        sub_directory = f'{TOPOLOGY_NAME}_{params.host}'
    elif params.exclude_sites is not None:
        ls_connection.filter_sites(include=False, sites=params.exclude_sites)
        sites_str: str = ', '.join(params.exclude_sites)
        details.append(f'Site(s) excluded: {sites_str}')
        sub_directory = f'{TOPOLOGY_NAME}_{params.host}'

    bgp_anchor = get_anchor(params.bgp_anchor)
    emblem_as = get_emblem(params.bgp_emblem_as)
    emblem_id = get_emblem(params.bgp_emblem_id)

    query: str = (
        'GET services\n'
        'Columns: host_name state description long_plugin_output\n'
        'Filter: description ~ BGP peer\n'
        'OutputFormat: python3\n'
    )
    if (raw_bgp_peers := ls_connection.query(query=query)) is not None:
        bgp_peers: MutableSequence[BgpPeer] = [
            BgpPeer.parse(host, state, service, raw_peer_data) for host, state, service, raw_peer_data in raw_bgp_peers
        ]

        # create index by local address, add host and service objects
        bgp_peers_by_local_addr: MutableMapping[str, MutableSequence[BgpPeer]] = {'0.0.0.0': []}
        for bgp_peer in bgp_peers:
            objects.add_host(bgp_peer.host)
            objects.add_service(bgp_peer.host, bgp_peer.service)
            connections.add_connection(
                right=bgp_peer.host,
                left=get_service(bgp_peer.host, bgp_peer.service),
                left_state=bgp_peer.state
            )

            if bgp_peer.local_address is not None:
                if bgp_peers_by_local_addr.get(bgp_peer.local_address) is None:
                    bgp_peers_by_local_addr[bgp_peer.local_address]: MutableSequence[BgpPeer] = []
                bgp_peers_by_local_addr[bgp_peer.local_address].append(bgp_peer)
            else:
                bgp_peers_by_local_addr['0.0.0.0'].append(bgp_peer)

        # find connections
        for bgp_peer in bgp_peers:
            if (peer_list := bgp_peers_by_local_addr.get(bgp_peer.remote_address)) is not None:
                for peer in peer_list:
                    if bgp_peer.local_as == peer.remote_as and bgp_peer.local_as is not None:
                        if bgp_peer.local_id == peer.remote_id and bgp_peer.local_id is not None:
                            if bgp_peer.local_address == peer.remote_address and bgp_peer.local_address is not None:
                                connections.add_connection(
                                    left=get_service(bgp_peer.host, bgp_peer.service),
                                    right=get_service(peer.host, peer.service),
                                    right_state=bgp_peer.state,
                                    left_state=peer.state,
                                )

            # if there is no peer_list the peer is either external (to checkmk) or the connection is not
            # established, so we have no local address or remote id
            else:
                ext_bgp_host_as: str | None = None
                ext_bgp_host_id: str | None = None
                if bgp_anchor in [
                    PARAM_BGP_EXT_ANCHOR_BOTH,
                    PARAM_BGP_EXT_ANCHOR_AS,
                ] and bgp_peer.remote_as is not None:
                    ext_bgp_host_as: str = f'BGP-AS: {bgp_peer.remote_as}'
                    objects.add_host(
                        host=ext_bgp_host_as,
                        link2core=False,
                        emblem=emblem_as,
                    )
                if bgp_anchor in [
                    PARAM_BGP_EXT_ANCHOR_BOTH,
                    PARAM_BGP_EXT_ANCHOR_ID
                ] and bgp_peer.remote_id is not None:
                    ext_bgp_host_id: str = f'BGP-ID: {bgp_peer.remote_id}'
                    objects.add_host(
                        host=ext_bgp_host_id,
                        link2core=False,
                        emblem=emblem_id
                    )
                    connections.add_connection(
                        right=ext_bgp_host_id,
                        left=get_service(bgp_peer.host, bgp_peer.service),
                        left_state=bgp_peer.state
                    )
                    if ext_bgp_host_as is not None:
                        connections.add_connection(ext_bgp_host_id, ext_bgp_host_as)

                if ext_bgp_host_as is not None and ext_bgp_host_id is None:
                    connections.add_connection(
                        right=get_service(bgp_peer.host, bgp_peer.service),
                        left=ext_bgp_host_as,
                        right_state=bgp_peer.state,
                    )

    connections.topo_connections.sort()
    data_set = {
        'version': 1,
        'name':  TOPOLOGY_NAME,
        'objects': dict(sorted(objects.topo_objects.items())),
        'connections': connections.topo_connections,
    }

    save_topology(data=data_set, sub_directory=sub_directory)
    # workaround for backend is only picking up topologies from default folder
    add_dummy_topologies(sub_directory=sub_directory)
    # end workaround
    make_topo_default(sub_directory=sub_directory, make_default=params.make_default)

    summary.append(f'Objects: {len(objects.topo_objects)}')
    details.append(f'Objects: {len(objects.topo_objects)}')

    summary.append(f'Connections: {len(connections.topo_connections)}')
    details.append(f'Connections: {len(connections.topo_connections)}')

    details.append(f'Written to: {BASE_TOPO_PATH}/{sub_directory}/data_{TOPOLOGY_NAME.lower()}.json')

    value = (time_ns() - start_time) / 1e9
    summary.append(f'Time taken: {render.timespan(value)}')
    details.append(f'Time taken: {render.timespan(value)}')

    perf_data = Metric(
        name=METRIC_TIME_TAKEN,
        value=value,
        # levels=None,
        # boundaries=None,
    )

    all_summary: str = ', '.join(summary)
    all_details: str = '\n'.join(details)
    print(f'{all_summary}\n{all_details}|{perf_data}')
    return 0


def parse_arguments(argv: Sequence[str]) -> Params:
    parser = ArgumentParser(
        prog='bgp_topology',
        formatter_class=RawTextHelpFormatter,
        description=f"""Create BGP peer network topology for Checkmk""",
        epilog=f"""
Example usage:
{OMD_ROOT}/{__USAGE__}        

Version: {__VERSION__} | Written by {__AUTHOR__}
for more information see: {__URL__}
        """
    )
    parser.add_argument(
        ARG_PARSER_ANCHOR,
        choices=[
            ARG_PARSER_ANCHOR_BOTH,
            ARG_PARSER_ANCHOR_AS,
            ARG_PARSER_ANCHOR_ID,
            ARG_PARSER_NONE,
        ],
        default=ARG_PARSER_ANCHOR_BOTH,
        help='Anchor for external BGP objects (default: %(default)s).',
    )
    parser.add_argument(
        ARG_PARSER_EMBLEM_AS,
        type=str,
        default=EMBLEM_BGP_AS,
        help='Emblem to use for BGP-AS objects (default: %(default)s).',
    )
    parser.add_argument(
        ARG_PARSER_EMBLEM_ID,
        type=str,
        default=EMBLEM_BGP_ID,
        help='Emblem to use for BGP-ID objects (default: %(default)s).',
    )
    parser.add_argument(
        ARG_PARSER_MAKE_DEFAULT,
        action='store_const', const=True,
        default=False,
        help='Make this topology the default (default: %(default)s).',
    )
    parser.add_argument(
        ARG_PARSER_HOST,
        type=str,
        help="""The name of the Checkmk host to which the plugin is attached. This is set 
automatically by Checkmk. If a site filter is active, the host name is 
appended to the subdirectory where the topology is stored (“BGP” becomes 
“BGP_host_name”). This way we can have more than one BGP topology without 
overwriting each other.""",
    )
    site_filter = parser.add_mutually_exclusive_group()
    site_filter.add_argument(
        ARG_PARSER_SITES_INCLUDE,
        type=str,
        nargs='+',
        help=f"""List of Checkmk site names to include in the topology creation. 
Can not used together with {ARG_PARSER_SITES_EXCLUDE}""",
    )
    site_filter.add_argument(
        ARG_PARSER_SITES_EXCLUDE,
        type=str,
        nargs='+',
        help=f"""List of Checkmk site names to exclude from the topology creation. 
Can not used together with {ARG_PARSER_SITES_INCLUDE}""",
    )

    return parser.parse_args(argv)


def main(argv: Sequence[str] | None = None) -> int:
    return create_bgp_topology(parse_arguments(argv=argv))
